// WindowOpen_ESP32C3_BLEtoWebServer.ino
// scans for window open BLE modules and makes status available to http connection
// only save BLE devices whose names have been pre-registered

// >>>> Using ESP32 board support V2.0.6 (V3.0.3 does not work) and board XIAO-ESP32C3  !!!!!!!!!!!!!!!
// Needs SafeString and ESPAutoWiFiConfig libraries installed.

// Needs Huge APP, 3M APP, 1M SPIFF  Partition setting
/*
  WindowOpen_ESP32C3_BLEtoWebServer.ino
   by Matthew Ford,  2024/08/05
   (c)2021 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.

*/

// Needs Huge APP, 3M APP, 1M SPIFF  Partition setting

#include <BLEDevice.h>
#include <BLEUtils.h>
#include <BLEScan.h>
#include <BLEAdvertisedDevice.h>

#include <WiFiClient.h>
#include <WebServer.h>
#include "ESP32_WS2812Flasher.h"
#include "ESPAutoWiFiConfig.h"
#include "ESPBufferedClient.h"

//#define DEBUG
static Stream *debugPtr = NULL;

const size_t MAX_BLE_ADVERTISED_NAME = 32;
// set your BLE device names here (without the ,C or ,O)
char expectedDevices[][MAX_BLE_ADVERTISED_NAME]
  = { "Front_Bedroom_Win_Rhs", "Front_Bedroom_Win_Lhs", "Front_Bathroom_Win", "Back_Bedroom_Win_Lhs", "Back_Bedroom_Win_Rhs" };

const size_t NO_EXPECTED_DEVICES = sizeof(expectedDevices) / MAX_BLE_ADVERTISED_NAME; // calculate the no of expected devices
typedef enum WINDOW_STATE_ENUM { CLOSED = 0, OPEN = 1, BAD = 2} windowStateEnum;

WebServer server(80);
void startWebServer();


#include "pfodLinkedPointerList.h" // iterable linked list of pointers to objects
#include "LastSeenOpenClosed.h"  // class for storing on linked list
#include "SafeString.h"
#include "millisDelay.h"
#include <WiFi.h>


millisDelay rebootDelay;
unsigned long REBOOT_DELAY_MS = 24ul * 60 * 60 * 1000; // 1day in mS

bool windowIsOpen = false;
bool isWindowOpen() {
  return windowIsOpen;
}

pfodLinkedPointerList<LastSeenOpenClosed> listOfLastSeen;
void processAdvertisedBLENames(SafeString &sfName, bool initFlag);

void initExpectedBLEDevices() {
  if (debugPtr) {
    debugPtr->println("Processing Expected Devices -------");
  }
  cSF(sfName, MAX_BLE_ADVERTISED_NAME); // max 32
  for (size_t i = 0; i < NO_EXPECTED_DEVICES; i++) {
    sfName.readFrom(expectedDevices[i], MAX_BLE_ADVERTISED_NAME - 2);
    sfName += ",C";
    processAdvertisedBLENames(sfName, true); // init
    sfName.clear(); // for next loop
  }
  if (debugPtr) {
    debugPtr->println("-------- End of Processing Expected Devices");
  }
}

// returns BAD if not found
windowStateEnum getWindowState(const char* deviceName) {
  LastSeenOpenClosed *devicePtr = listOfLastSeen.getFirst();
  while (devicePtr) {
    if (strcmp(deviceName, devicePtr->getDeviceName()) == 0) { // match
      float lastSeen_sec = (millis() - devicePtr->getLastSeen()) / 1000.0;
      if (lastSeen_sec > 220) {
        return BAD;
      }
      // else
      if (devicePtr->getIsOpen()) {
        return OPEN;
      } //else
      return CLOSED;
    } // else
    devicePtr = listOfLastSeen.getNext();
  }
  return BAD; // null not found
}

// returns NULL if not found
LastSeenOpenClosed* getLastSeenOpenClosed(SafeString &name) {
  LastSeenOpenClosed *devicePtr = listOfLastSeen.getFirst();
  while (devicePtr) {
    if (name == (devicePtr->getDeviceName())) {
      return devicePtr;
    } // else
    devicePtr = listOfLastSeen.getNext();
  }
  return devicePtr; // null not found
}

int scanTime = 2; //In seconds
BLEScan *pBLEScan;

// only add devices if initFlag is true, else skip those not already there
void processAdvertisedBLENames(SafeString &sfName, bool initFlag) {
  if (debugPtr) {
    debugPtr->print("Device name: ");
    debugPtr->println(sfName);
  }
  // only scan for == upto first ,
  // look at names with trailing ,O or ,C
  if ((!sfName.endsWith(",O")) && (!sfName.endsWith(",C"))) {
    return; // skip this one
  } // else
  bool windowOpen = sfName.endsWith(",O");

  int idx = sfName.indexOf(',');
  if (idx >= 0) {
    sfName.substring(sfName, 0, idx);
  }

  LastSeenOpenClosed *devicePtr = getLastSeenOpenClosed(sfName);
  if ((!devicePtr) && (initFlag)) { // not found
    // else if initializing add
    // not  found add it upto first ,
    if (debugPtr) {
      debugPtr->print("Adding Device name: "); debugPtr->println(sfName);
    }
    devicePtr = new LastSeenOpenClosed(sfName.c_str()); // note MUST use new since pfodLinkedPointerList uses delete when remove() called
    listOfLastSeen.add(devicePtr);
  } else if (devicePtr) {
    // update lastseen
    devicePtr->updateLastSeen(millis());
    devicePtr->setIsOpen(windowOpen); // save window state
  }
}

class MyAdvertisedDeviceCallbacks : public BLEAdvertisedDeviceCallbacks {
    void onResult(BLEAdvertisedDevice advertisedDevice) {
      if (advertisedDevice.haveName()) {
        cSF(sfName, MAX_BLE_ADVERTISED_NAME); // max 32
        sfName.readFrom(advertisedDevice.getName().c_str());
        processAdvertisedBLENames(sfName, false); // skip if not already there
      }
    };
};

static TaskHandle_t bleScannerHandle = NULL;


void BLE_init() {
  BLEDevice::init("");
  pBLEScan = BLEDevice::getScan(); //create new scan
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setActiveScan(true);
  pBLEScan->setInterval(40);  // same as Apple in foreground
  pBLEScan->setWindow(30); // same as Apple in foreground, less or equal setInterval value
}

/* this function will be invoked when additionalTask was created */
void bleScannerTask( void * parameter ) {
  BLE_init();
  /* loop forever */
  for (;;) {
    // put your taskain code here, to run repeatedly:
    pBLEScan->clearResults(); // delete results fromBLEScan buffer to release memory
    BLEScanResults foundDevices = pBLEScan->start(scanTime, false);
    if (debugPtr) {
      debugPtr->print("Devices found: ");
      debugPtr->println(foundDevices.getCount());
    }
    delay(1000); // give other things a change to run
  }
  /* delete a task when finish,
    this will never happen because this is infinity loop */
  vTaskDelete( NULL );
}

void setUpWiFiServices() {
  startWebServer();
}

// returns true if window open Or badDevice
bool checkWindowOpen() {
  LastSeenOpenClosed *devicePtr = listOfLastSeen.getFirst();
  while (devicePtr != NULL) {
    float lastSeen_sec = (millis() - devicePtr->getLastSeen()) / 1000.0;
    if (lastSeen_sec > 220) {
      devicePtr = listOfLastSeen.getNext();
      continue; // lost connection, skip this one
    }
    if (devicePtr->getIsOpen()) { // open and active not missed two advert interval 10+90 x 2
      windowIsOpen = true;
      return true; // window open
    }
    devicePtr = listOfLastSeen.getNext();
  }
  windowIsOpen = false;
  return false;
}

bool checkBadWindow() {
  LastSeenOpenClosed *devicePtr = listOfLastSeen.getFirst();
  while (devicePtr != NULL) {
    float lastSeen_sec = (millis() - devicePtr->getLastSeen()) / 1000.0;
    if (lastSeen_sec > 220) {
      return true; // lost connection
    }
    devicePtr = listOfLastSeen.getNext();
  }
  return false;
}

void setup() {
  Serial.begin(115200);
#ifdef DEBUG
  for (int i = 10; i > 0; i--) {
    Serial.print(i); Serial.print(' ');
    delay(500);
  }
  SafeString::setOutput(Serial); // for debug error msgs
  debugPtr = &Serial;
  debugPtr->println();
#endif

  //  setESPAutoWiFiConfigDebugOut(Serial); // note this prints out network password!!
  if (ESPAutoWiFiConfigSetup(0, true, 0)) {
    return;
  }

  // NOTE: do not have this code before ESPAutoWiFiConfigSetup as it will cause the XIAO ESP32C3 to reboot on watch dog, ESP32 is OK
  rebootDelay.start(REBOOT_DELAY_MS); // start reboot timer
  enableLoopWDT(); // default appears to be 5sec
  initExpectedBLEDevices(); //  do this before setUpWiFiServices
  setUpWiFiServices();
  BaseType_t err = xTaskCreate(
                     bleScannerTask, /* Task function. */
                     "bleScannerTask", /* name of task. */
                     102400, /* Stack size of task */
                     NULL, /* parameter of the task */
                     1, /* priority of the task */ // the loop task also has priority 1 also so will share
                     &bleScannerHandle ); /* Task handle to keep track of created task */
  if (((int)err) != 1) { // pPass ,  pdTRUE      ( ( BaseType_t ) 1 ) in #include "freertos/projdefs.h"
    if (debugPtr) {
      debugPtr->print("xTaskCreatePinnedToCore returned:");
      debugPtr->println((int)err);
    }
  }
  setUpWiFiServices();
}


void loop() {
  if (ESPAutoWiFiConfigLoop()) {  // handle auto configure/connect
    return;
  }
  // normal loop
  if (rebootDelay.justFinished()) {
    while (1) {} // force watch dog timer reboot
  }

  server.handleClient();
  yield();
}


// ========== trivial webserver example =============
void notFound() {
  server.send(404, "text/plain", "Not found");
}

cSF(sfMsg, 1024);
void sendWindowStatus() {
  if (debugPtr) {
    debugPtr->println(">>> WebServer sendWindowStatus");
  }
  sfMsg.clear();
  LastSeenOpenClosed *devicePtr = listOfLastSeen.getFirst();
  while (devicePtr != NULL) {
    const char *deviceName = devicePtr->getDeviceName();
    sfMsg.print(deviceName);
    sfMsg.print(',');
    float lastSeen_sec = (millis() - devicePtr->getLastSeen()) / 1000.0;
    if (lastSeen_sec > 220) { // missed two advert interval 10+90 x 2
      sfMsg.println('B'); // bad
    } else {
      if (devicePtr->getIsOpen()) {
        sfMsg.println('O');
      } else {
        sfMsg.println('C');
      }
    }
    devicePtr = listOfLastSeen.getNext();
  }
  server.send(200, "text/plain", sfMsg.c_str());
}

void handleRoot() {
  if (debugPtr) {
    debugPtr->println(">>> WebServer handleRoot");
  }
  String msg =
    "<html>\
  <head>\
    <meta http-equiv='refresh' content='5'/>\
    <title>Window Open Monitor</title>\
    <style>\
      body { background-color: #cccccc; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }\
    </style>\
  </head>\
  <body>";
  msg += "<br>";
  msg += "<br>";
  msg += "The BLE Window Open devices found were:-<br>";
  // loop through and add reading found
  LastSeenOpenClosed *devicePtr = listOfLastSeen.getFirst();
  //  msg += "<h1>";
  if (devicePtr == NULL) {
    msg += "No devices found so far.<br>";
  } else {
    while (devicePtr != NULL) {
      const char *deviceName = devicePtr->getDeviceName();
      float lastSeen_sec = (millis() - devicePtr->getLastSeen()) / 1000.0;
      if (lastSeen_sec > 220) { // missed two advert interval 10+90 x 2
        msg += "<b><font size=\"+1\" color=\"black\">";
        msg += " BAD    -- ";
        msg += deviceName;
        msg += " - not advertising. Battery Low or Out Of Range</font></b>";
      } else {
        if (devicePtr->getIsOpen()) {
          msg += "<b><font size=\"+2\" color=\"red\">";
          msg += " OPEN   -- ";
          msg += deviceName;
          msg += "</font></b>";
        } else {
          msg += "<font color=\"green\">";
          msg += " CLOSED -- ";
          msg += deviceName;
          msg += "</font>";
        }
      }
      msg += "<font size=\"-1\">";
      msg += " &nbsp;&nbsp;&nbsp;";
      msg += (millis() - devicePtr->getLastSeen()) / 1000.0;
      msg += " sec ago";
      msg += "</font>";
      msg += "<br>";
      devicePtr = listOfLastSeen.getNext();
    }
  }
  //  msg += "</h1>";
  msg += "</body></html>";

  server.send(200, "text/html", msg);
}

void startWebServer() {
  server.on("/", handleRoot);
  server.on("/status", sendWindowStatus);
  server.onNotFound(notFound);
  server.begin();
  if (debugPtr) {
    debugPtr->println("webserver started");
  }
}
